@@ -5,8 +5,10 @@
 #
 
 # Network output to box
-from ._cornernet import *
-from ._darknet import *
+from ._anchor_yolo import *
+from ._anchor_oriented import *
+from ._anchor_masked import *
+from ._corner import *
 
 # Util
 from ._brambox import *
