@@ -7,4 +7,4 @@
 
 from ._fit import *
 from ._augment import *
-from ._util import *
+from ._base import *
