@@ -5,5 +5,7 @@ It is mostly targeted at Object detection networks, but contains the necessary t
 """
 
 from . import layer
+from . import backbone
+from . import head
 from . import loss
 from . import module
